#include <stddef.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <stdlib.h>
#include <libgreattao/tao.h>
#include "dir.h"

struct copy_operation {

  DIR *dir;
  const char *dst, *src;
  struct copy_operation *parent;
};

struct remove_operation {

  DIR *dir;
  const char *src;
  struct remove_operation *parent;
};

void *main_window, *progress_window;
void *operation_timer;
struct view *active_view;
int source, destination;
DIR *current_dir;

void do_copy_directory(struct copy_operation *cp);
void copy(const char *source_, const char *destination_, struct copy_operation *pcp);
void do_file_remove(const char *location, struct remove_operation *prem);

void set_active_view(struct view *curr)
{
  active_view = curr;
}

void *get_main_window(void)
{
  return main_window;
}

void app_callback(void *mesh, void *data)
{
  if ((int)(intptr_t) data == 0) {
    
    tao_close();
  }
}


void close_message(void *mesh, void *window)
{
  tao_release_window(window);
}

void create_message_window(const char *message, const char *code)
{
  void *message_window;
  
  message_window = tao_new_window("/desktop/dialogs/message_dialog");
  tao_add_handler(message_window, "/message", NULL, 0);
  tao_add_handler(message_window, "/code", NULL, 0);
  tao_set_hint(message_window, "/code", 0, (char*)code);
  tao_set_hint(message_window, "/message", 0, (char*)message);
  tao_add_handler(message_window, "/:abort", close_message, message_window);
}

void set_dir(void *mesh, struct view *view)
{
   char *path;
   int error;
   
  path = tao_get_string(main_window, (char*)view->tao_edit, &error);
  if (error || !path) {
  
    create_message_window("Error while getting path. Is it not empty?", "2");
    return;
  }
  
  if (path[0] != '/') {
  
    create_message_window("Path must start with /", "3");
    return;
  }
  
  if (!load_directory(view, path)) {
  
    create_message_window("Unable to open directory", "1");
  }
  free(path);
}

void copy_helper(struct copy_operation *data)
{
  static char buffer[512];
  static int file_size = 0;
  static int copied_bytes = 0;
  int count;
  
  if (file_size == 0) {
  
    file_size = lseek(source, 0, SEEK_END);
    lseek(source, 0, SEEK_SET);
  }
  
  count = read(source, buffer, 512);
  if (!count || count == -1) {
    
    tao_remove_timer(operation_timer);
    operation_timer = NULL;
    
    close(source);
    close(destination);

    tao_release_window(progress_window);
    progress_window = NULL;
    file_size = 0;
    copied_bytes = 0;
    load_directory(&left_view, left_view.dir_path);
    load_directory(&right_view, right_view.dir_path);
    
    if (data) {
      
      do_copy_directory(data);
    }
    return;
  }
  write(destination, buffer, count);
  
  copied_bytes += count;
  snprintf(buffer, 512, "%d", copied_bytes * 100 / file_size);
  tao_set_hint(progress_window, "/progress/main", 0, buffer);
}

void copy_file(const char *sourcep, const char *destinationp,  int source_, int destination_, struct copy_operation *cp)
{
  int length;
  char *buffer;
  
  source = source_;
  destination = destination_;
  
  operation_timer = tao_add_timer(1, (void (*)(void*))copy_helper, cp);
  progress_window = tao_new_window("/desktop/dialogs/progress_dialog");
  tao_add_handler(progress_window, "/progress/main", NULL, 0);
  tao_set_modal(main_window, "/progress_dialog", progress_window, 1);

  length = strlen(sourcep)  + strlen(destinationp) + strlen("Copying from \"\" to \"\"");
  buffer = malloc(length + 1);
  
  snprintf(buffer, length+1, "Copying from \"%s\" to \"%s\"", sourcep, destinationp);
  
  tao_add_handler(progress_window, "/progress/main/current task", NULL, 0);
  tao_set_hint(progress_window, "/progress/main/current task", 1, buffer);
  free(buffer);
  
  ftruncate(destination_, 0);
}

void do_copy_directory(struct copy_operation *cp)
{
  struct dirent *dirent;
  int sfd, dft;
  char *buffer;
  char *buffer2;
  int length;
  
  if (cp == NULL) {
  
    return;
  }
  
  while (1) {
    
    dirent = readdir(cp->dir);
    if (!dirent) {
    
      free((void*)cp->src);
      free((void*)cp->dst);
      closedir(cp->dir);
      do_copy_directory(cp->parent);
      free(cp);
      return;
    }
    
    if (strcmp(dirent->d_name, ".") != 0 &&
      strcmp(dirent->d_name, "..") != 0) {
    
	break;
      }
  }
  length = strlen(cp->src) + strlen(dirent->d_name) + 1;
  buffer = malloc(length + 1);
  strcpy(buffer, cp->src);
  strcat(buffer, "/");
  strcat(buffer, dirent->d_name);
  
  length = strlen(cp->dst) + strlen(dirent->d_name) + 1;
  buffer2 = malloc(length + 1);
  strcpy(buffer2, cp->dst);
  strcat(buffer2, "/");
  strcat(buffer2, dirent->d_name);
  
  copy(buffer, buffer2, cp);
}

void copy(const char *source_, const char *destination_, struct copy_operation *pcp)
{
  struct stat src, dst;
  struct copy_operation *cp;
  int sfd, dfd;
  
  if (stat(source_, &src) == -1) {
    
    create_message_window("Unable to check type of source file", "7");
    free((void*)source_);
    free((void*)destination_);
    return;
  }
  
  if ((src.st_mode & S_IFMT) == S_IFDIR) {
    
    if (stat(destination_, &dst) == -1) {
      
      if (mkdir(destination_, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH) == -1) {
	
	create_message_window("Unable to create directory", "6");
        free((void*)source_);
	free((void*)destination_);
	return;
      }
   }
    
    cp = (struct copy_operation*) malloc(sizeof(*cp));
    cp->dir = opendir(source_);
    cp->src = strdup(source_);
    cp->dst = strdup(destination_);
    cp->parent = pcp;
    
    do_copy_directory(cp);
  }
  else if ((src.st_mode & S_IFMT) == S_IFREG) {
    
    sfd = open(source_, O_RDONLY);
    if (sfd == -1) {
    
      create_message_window("Unable to open/create source file", "5");
      free((void*)source_);
      free((void*)destination_);
      return;
    }
    
    dfd = open(destination_, O_WRONLY | O_CREAT);
    if (dfd == -1) {
    
      close(sfd);
      create_message_window("Unable to open/create destination file", "4");
      free((void*)source_);
      free((void*)destination_);
      return;
    }
    
    copy_file(source_, destination_, sfd, dfd, pcp);
    
    free((void*)source_);
    free((void*)destination_);
  }
}

void do_remove_directory(struct remove_operation *rem)
{
  char *buffer;
  int length;
  struct dirent *dirent;
  
  if (!rem) {
    
    return;
  }
  
  while (1) {
    dirent = readdir(rem->dir);
  
    if (!dirent) {
 
    rmdir(rem->src);
    free((void*)rem->src);
    do_remove_directory(rem->parent);
    free(rem);
    
    return;
  }
  
  if (strcmp(dirent->d_name, ".") != 0 &&
      strcmp(dirent->d_name, "..") != 0) {
    
    break;
      }
  }
  
  length = strlen(rem->src) + strlen(dirent->d_name) + 1;
  buffer = (char*) malloc(length+1);
  
  strcpy(buffer, rem->src);
  strcat(buffer, "/");
  strcat(buffer, dirent->d_name);
  
  do_file_remove(buffer, rem);
}

void do_file_remove(const char *location, struct remove_operation *prem)
{
  struct stat src;
  struct remove_operation *rem;
  
  if (stat(location, &src) == -1) {
   
    return;
  }
  
  if ((src.st_mode & S_IFMT) == S_IFDIR) {
    
    rem = (struct remove_operation*) malloc(sizeof(*rem));
    
    rem->parent = prem;
    rem->src = strdup(location);
    rem->dir = opendir(location);
    
    do_remove_directory(rem);
  }
  else if ((src.st_mode & S_IFMT) == S_IFREG) {
    
    unlink(location);
    do_remove_directory(prem);
  }
}

void file_callback(void *mesh, void *data)
{
  struct stat  file_stat;
  struct view *no_active_view;
  char *file_path, *file_path2;
  int length;
  int fd, fd2;
  int inode, device;
  char *current, *last_slash;
  char *to_set;
  
  if (!active_view) {
    
    return;
  }
  
  if (active_view == &left_view) {
    
      no_active_view = &right_view;
    }
    else {
      
      no_active_view = &left_view;
  }
  
  if (active_view->selected) {
    
    length = strlen(active_view->dir_path) + strlen(active_view->selected->name) + 1;
    file_path = (char *) malloc(length + 1);
    strcpy(file_path, active_view->dir_path);
    strcat(file_path, "/");
    strcat(file_path, active_view->selected->name);
  }
  else {
    
    length = strlen(active_view->dir_path);
    file_path = (char *) malloc(length + 1);
    strcpy(file_path, active_view->dir_path);
  }
  
  to_set = strdup(active_view->dir_path);
  
  if ((int)(intptr_t) data == 0) {
    
    stat(file_path, &file_stat);
    
    if ((file_stat.st_mode & S_IFMT) == S_IFDIR) {
    
        free(to_set);
        to_set = strdup(file_path);
    }
    else if ((file_stat.st_mode & S_IFMT) == S_IFREG) {
    
      if (!fork()) {
	execlp("xdg-open", "xdg-open", file_path, NULL);
      }
    }
  }
  else if ((int)(intptr_t) data == 1) {
    
    if (no_active_view->selected) {
    
      length = strlen(no_active_view->dir_path) + strlen(no_active_view->selected->name) + 1;
      file_path2 = (char *) malloc(length + 1);
      strcpy(file_path2, no_active_view->dir_path);
      strcat(file_path2, "/");
      strcat(file_path2, no_active_view->selected->name);
    }
    else {
      
      length = strlen(no_active_view->dir_path) + strlen(active_view->selected->name) + 1; 
      file_path2 = (char *) malloc(length + 1);
      strcpy(file_path2, no_active_view->dir_path);
      strcat(file_path2, "/");
      strcat(file_path2, active_view->selected->name);
    }
    
    if (stat(file_path, &file_stat) == -1) {
    
      free(file_path);
      free(file_path2);
      create_message_window("Unable to check type of source file", "8");
      return;
    }
    
    device = file_stat.st_dev;
    inode  = file_stat.st_ino;
    
    
    if (stat(file_path2, &file_stat) == 0 && device == file_stat.st_dev && inode == file_stat.st_ino)  {
    
      free(file_path);
      free(file_path2);
      create_message_window("You cannot copy file to the same file", "10");
      return;
    }
    
    copy(strdup(file_path), strdup(file_path2), 0);
    free(file_path2);
    
  }
  else if ((int)(intptr_t) data == 2) {
    
    do_file_remove(file_path, NULL);
  }
  else if ((int)(intptr_t) data == 3) {
  
    free(file_path);
    file_path = strdup(active_view->dir_path);
    current = file_path;
    last_slash = NULL;
    
    /* Omit first slash */
    ++current;
    while (*current != 0) {
      
      if (*current == '/' && current[1] != 0) {
      
	last_slash = current;
      }
      ++current;
    }
    
    free(to_set);
    if (last_slash) {
    
      *last_slash = 0;
      to_set = strdup(file_path);
    }
    else {
      
      to_set = strdup("/");
    }
  }
  

  if (strcmp(active_view->dir_path, no_active_view->dir_path) == 0) {
  
    //load_directory(no_active_view, no_active_view->dir_path);
  }
 
  if (load_directory(active_view, to_set)) {
  
    tao_set_string(main_window, (char*)active_view->tao_edit, to_set);
  }
  else {
    create_message_window("Unable to open directory", "1");
  }

  free(file_path);
  free(to_set);
}

int main(int argc, char **argv)
{
  char *cwd;
  
  tao_initialize("tao-manager", "", &argc, argv);
  
  main_window = tao_new_window("/app/file_manager");
  tao_add_handler(main_window, "/:abort", app_callback, (void*)0);
  
  //tao_add_handler(main_window, "/actions/app", NULL, 0);
  tao_add_handler(main_window, "/actions/app/close", app_callback, (void*)0);
  
  //tao_add_handler(main_window, "/actions/file", NULL, 0);
  tao_add_handler(main_window, "/actions/file/open", file_callback, (void*)0);
  tao_add_handler(main_window, "/actions/file/copy", file_callback, (void*)1);
  tao_add_handler(main_window, "/actions/file/remove", file_callback, (void*)2);
  tao_add_handler(main_window, "/actions/file/dirup", file_callback, (void*)3);
  
  //tao_add_handler(main_window, "/fileviews/left", NULL, 0);
  //tao_add_handler(main_window, "/fileviews/right", NULL, 0);
  
  tao_add_handler(main_window, "/fileviews/left/path_edit", set_dir, &left_view);
  tao_add_handler(main_window, "/fileviews/right/path_edit", set_dir, &right_view);
  
  tao_add_handler(main_window, "/fileviews/left/list/columns/name", NULL, 0);
  tao_add_handler(main_window, "/fileviews/right/list/columns/name", NULL, 0);
  
  tao_set_hint(main_window, "/fileviews/left/list/columns/name", 0, "File name");
  tao_set_hint(main_window, "/fileviews/right/list/columns/name", 0, "File name");
  
  
  tao_add_handler(main_window, "/fileviews/left/list/columns/size", NULL, 0);
  tao_add_handler(main_window, "/fileviews/right/list/columns/size", NULL, 0);
  
  if (argc > 1) {
  
    cwd = argv[1];
  }
  else {
  
    cwd = getcwd(NULL, 0);
  }
  left_view.tao_path = "/fileviews/left";
  right_view.tao_path = "/fileviews/right";
  left_view.tao_edit = "/fileviews/left/path_edit";
  right_view.tao_edit ="/fileviews/right/path_edit";
  
  load_directory(&left_view, cwd);
  load_directory(&right_view, cwd);
  
  tao_set_string(main_window, (void*)left_view.tao_edit, cwd);
  tao_set_string(main_window, (void*)right_view.tao_edit, cwd);
  
  tao_handle_events();
}
